"""
**Functional interface to the plot window manager**

The interface is very much MATLAB-like. 

If talkback is enabled the return values of command handlers are returned by 
the functions. If talkback is disabled, the return values are always ``None``. 

The plotting happens in a separate thread (process) so the plot windows remain 
responsive even if the main thread does some computation. No additional user 
intervention is needed (like calling some refresh function from time-to time). 

An example will explain more than ten paragraphs of text::
	
	import pyopus.wxmplplot.plotitf as pyopl
	from numpy import arange, sin, cos, exp, pi
	
	# Initialize the graphical thread. If it is already initialized this does 
	# nothing. After initialization a Control window pops up. When it is closed 
	# all plot windows are closed and the graphical thread exits. After that a 
	# new call to init() is needed to start the graphical thread. 
	pyopl.init()
	
	# Create some x and y data
	x = arange(0.0, 2*pi, 0.2)
	y1 = sin(x)
	y2 = cos(x)
	y3 = exp(x/pi)
	
	# Create a plot window (figure). This is now the active plot window. 
	# Tag (stored in f1) is assigned automatically by the system. 
	# Figure size is 600x400 pixels. The dpi value is used when the figure is 
	# saved to Postscript. 
	f1=pyopl.figure(None, windowTitle="Figure - single axes", 
					figpx=(600,400), dpi=100)	
	
	# Create axes in the active figure. Axes will take the whole area of the 
	# active figure. The return value (stored in ax1) is the axes tag. 
	# The tag is generated by the system. 
	ax1=pyopl.axes()
	
	# Add first trace (x, y1). Solid line, points marked with o. 
	# Color is a (r,g,b) tuple of values between 0 and 1. The trace is labeled 
	# 'sin(x)'. See Matplotlib for other style options. 
	pyopl.plot(x, y1, '-o', label='sin(x)', color=(1,0,0))
	
	# Hold on - next traces will not delete previous ones. 
	pyopl.hold(True)
	
	# Add second trace. Red, points marked with x, no line. Trace is labeled 
	# 'cos(x)'. 
	pyopl.plot(x, y2, 'rx', label='cos(x)')
	
	# Add third trace. Dashed, black, points marked with vertical lines. 
	# Label is 'exp(x/pi)'. 
	pyopl.plot(x, y3, '--k|', label='exp(x/pi)')
	
	# Add legend. 
	pyopl.legend()
	
	# Label axes, add title. 
	pyopl.xlabel('x-axis')
	pyopl.ylabel('y-axis')
	pyopl.title('Axes title. ')
	
	# Turn on grid. 
	pyopl.grid(True)
	
	# Disable hold. We are done adding new traces. 
	pyopl.hold(False)
	
	# Enable active figure updates. 
	pyopl.repaintFigure(True)
"""

from wxmgr import *
from matplotlib import rcParams
import os

__all__ = [ 'init', 'shutdown', 'setTalkback', 'updateRCParams', 
			'plot', 'stem', 'semilogx', 'semilogy', 'loglog', 
			'setXAxisScale', 'setYAxisScale', 
			'grid', 'xgridline', 'ygridline', 'arrow', 
			'getXlim', 'getYlim', 'getRlim', 'setXlim', 'setYlim', 'setRlim', 'autoscale', 
			'setXTicks', 'setYTicks', 'getXTicks', 'getYTicks', 
			'legend', 'xlabel', 'ylabel', 'title', 'text', 'annotate', 
			'figure', 'setFigure', 'getFigure', 'getAllFigures', 'closeFigure', 'closeAllFigures', 'clearFigure', 
			'showFigure', 'raiseFigure', 'repaintFigure', 'saveFigure', 
			'figureTitle', 'windowTitle', 
			'axes', 'subplot', 'setAxes', 'getAxes', 'getAllAxes', 'deleteAxes', 'clearAxes', 'hold'
		]

# TODO: plot axrlines, axthetalines, get/set thetalim

GUIctl=None
"The :class:`GUIControl` object for communicationg with the graphical thread."	

def init():
	"""
	Initializes the graphical thread and pops up the Control window. Closing 
	the control window closes all plots and exits the graphical thread. To 
	use graphics again a new call to :func:`init` is needed. 
	
	If the graphical therad is already running calling :func:`init` has no 
	effect. 
	"""
	global GUIctl
	
	if GUIctl is None:
		GUIctl=GUIfactory()
		
	GUIctl.StartGUI()
	updateRCParams({
		'font.size': 10,
		'axes.titlesize': 10,
		'axes.labelsize': 9,
		'xtick.labelsize': 9,
		'ytick.labelsize': 9,
		'legend.fontsize': 9
	})
	GUIctl.setTalkback(True)

def shutdown():
	"""
	Exits the graphical thread. Equivalent to closing the Control window. 
	"""
	global GUIctl 
	
	if GUIctl is not None:
		GUIctl.StopGUI()

def updateRCParams(*args, **kwargs): 
	"""
	Calling convention: updateRCParams(*dict*)
	
	Updates the rcParams structure of Matplotlib with the dictionary given by 
	*dict*. Returns ``True`` on success. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'updatercparams'], 
			'args': args, 
			'kwargs': kwargs
		}
	)
		
def setTalkback(on=True):
	"""
	Enables (*on* is ``True``) or disables talkback. If talkback is disabled 
	all functions in this module always return ``None``. 
	"""
	global GUIctl
	
	GUIctl.setTalkback(on)
	
	
def plot(*args, **kwargs):
	"""
	Plots a trace on the active axes of the active plot window. 
	
	See the Matplotlib :meth:`Axes.plot` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('plot',) + args, 
			'kwargs': kwargs
		}
	)

def stem(*args, **kwargs):
	"""
	Plots a stem plot on the active axes of the active plot window. 
	
	See the Matplotlib :meth:`Axes.stem` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('stem',) + args, 
			'kwargs': kwargs
		}
	)

def semilogx(*args, **kwargs):
	"""
	Plots a trace on the active axes of the active plot window. 
	Makes the scale of the x-axis logarithmic. 
	
	See the Matplotlib :meth:`Axes.semilogx` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('semilogx',) + args, 
			'kwargs': kwargs
		}
	)

def semilogy(*args, **kwargs):
	"""
	Plots a trace on the active axes of the active plot window. 
	Makes the scale of the y-axis logarithmic. 
	
	See the Matplotlib :meth:`Axes.semilogy` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('semilogy',) + args, 
			'kwargs': kwargs
		}
	)

def loglog(*args, **kwargs):
	"""
	Plots a trace on the active axes of the active plot window. 
	Makes the scale of the x-axis and the y-axis logarithmic. 
	
	See the Matplotlib :meth:`Axes.loglog` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('loglog',) + args, 
			'kwargs': kwargs
		}
	)

def getXlim():
	"""
	Returns the x-axis range for the active axes of the active plot window. 
	The range is returned as a list with 2 members (low and high). 
	
	See the Matplotlib :meth:`Axes.get_xlim` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('get_xlim',), 
			'kwargs': {}
		}
	)
	
def getYlim():
	"""
	Returns the y-axis range for the active axes of the active plot window. 
	The range is returned as a list with 2 members (low and high). 
	
	See the Matplotlib :meth:`Axes.get_ylim` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('get_ylim',), 
			'kwargs': {}
		}
	)

def getRlim():
	"""
	Returns the radial range for the active axes of the active plot window. 
	Works only if axes are displaying a polar plot. 
	The range is returned as a real number (maximal R). 
	
	See the Matplotlib :meth:`PolarAxes.get_rmax` method for details. 
	"""
	# Get R range of polar axes. See axes.get_rmax()
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('get_rmax',), 
			'kwargs': {}
		}
	)
	
def setXlim(*args, **kwargs):
	"""
	Calling convention: setXlim(*xmin*, *xmax*)
	
	Sets the x-axis limits for the active axes of the active plot window. 
	The limits are specified with the *xmin* and the *xmax* argument. 
	Any of the two arguments can be ommitted. This way only the specified limit 
	is affected. 
	
	See the Matplotlib :meth:`Axes.set_xlim` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('set_xlim',) + args, 
			'kwargs': kwargs
		}
	)

def setYlim(*args, **kwargs):
	"""
	Calling convention: setYlim(*ymin*, *ymax*)
	
	Sets the y-axis limits for the active axes of the active plot window. 
	The limits are specified with the *ymin* and the *ymax* argument. 
	Any of the two arguments can be ommitted. This way only the specified limit 
	is affected. 
	
	See the Matplotlib :meth:`Axes.set_ylim` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('set_ylim',) + args, 
			'kwargs': kwargs
		}
	)

def setRlim(*args, **kwargs):
	"""
	Calling convention: setRlim(*rmax*)
	
	Sets the R limit for the active axes of the active plot window to *rmax*. 
	Works only if axes are displaying a polar plot. 
	
	See the Matplotlib :meth:`PolarAxes.set_rmax` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('set_rmax',) + args, 
			'kwargs': kwargs
		}
	)

def autoscale(*args, **kwargs):
	"""
	Calling convention: autoscale(*tight*, *scalex*, *scaley*)
	
	Autoscales the active axes of the active plot window. *scalex* and *scaley* 
	are boolean flags specifying which axes to autoscale. If *tight* is 
	``True`` autoscaling is tight. 
	
	See the Matplotlib :meth:`Axes.autoscale_view` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('autoscale_view',) + args, 
			'kwargs': kwargs
		}
	)

def setXTicks(*args, **kwargs):
	"""
	Calling convention: setXTicks(*ticks*, *minor*)
	
	Sets the list of ticks for the x-axis. *minor* is ``False`` by default. 
	
	See the Matplotlib :meth:`Axes.set_xticks` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('set_xticks',) + args, 
			'kwargs': kwargs
		}
	)

def setYTicks(*args, **kwargs):
	"""
	Calling convention: setYTicks(*ticks*, *minor*)
	
	Sets the list of ticks for the y-axis. *minor* is ``False`` by default. 
	
	See the Matplotlib :meth:`Axes.set_yticks` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('set_yticks',) + args, 
			'kwargs': kwargs
		}
	)

def getXTicks(*args, **kwargs):
	"""
	Calling convention: getXTicks(*minor*)
	
	Returns the list of ticks for the x-axis. *minor* is ``False`` by default. 
	
	See the Matplotlib :meth:`Axes.get_xticks` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('get_xticks',) + args, 
			'kwargs': kwargs
		}
	)

def getYTicks(*args, **kwargs):
	"""
	Calling convention: getYTicks(*minor*)
	
	Returns the list of ticks for the y-axis. *minor* is ``False`` by default. 
	
	See the Matplotlib :meth:`Axes.get_yticks` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('get_yticks',) + args, 
			'kwargs': kwargs
		}
	)
	
def grid(*args, **kwargs):
	"""
	Calling convention: grid(*b*, ...)
	
	Turns the grid on/off (depending on *b*) for the active axes of the active 
	plot window. All remaining arguments are passed to the 
	Matplotlib :meth:`Axes.grid` method. 
	
	See the Matplotlib :meth:`Axes.grid` method for details. 
	"""
	GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('grid',) + args, 
			'kwargs': kwargs
		}
	)

def xgridline(*args, **kwargs):
	"""
	Calling convention: xgridline(*x* =0.0, *ymin* =0.0, *ymax* =1.0, ...)
	
	Draws a vertical grid line at *x* from *ymin* to *ymax* in the active axes 
	of the active plot window. *ymin* and *ymax* are both in relative 
	coordinates (between 0 and 1, 0 for bottom, 1 for top). All remaining 
	arguments are passed to the Matplotlib :meth:`Axes.axvline` method. 
	
	See the Matplotlib :meth:`Axes.axvline` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('axvline',) + args, 
			'kwargs': kwargs
		}
	)

def ygridline(*args, **kwargs):
	"""
	Calling convention: ygridline(*y* =0.0, *xmin* =0.0, *xmax* =1.0, ...)
	
	Draws a vertical grid line at *y* from *xmin* to *xmax* in the active axes 
	of the active plot window. *xmin* and *xmax* are both in relative 
	coordinates (between 0 and 1, 0 for left, 1 for right). All remaining 
	arguments are passed to the Matplotlib :meth:`Axes.axhline` method. 
	
	See the Matplotlib :meth:`Axes.axhline` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('axhline',) + args, 
			'kwargs': kwargs
		}
	)
	
def arrow(*args, **kwargs):
	"""
	Calling convention: arrow(*x*, *y*, *dx*, *dy*, ...)
	
	Draws an arrow from (*x*, *y*) to (*x*+*dx*, *y*+*dy*) in the active axes 
	of the actigve plot window. All remaining arguments are passed to the 
	Matplotlib :meth:`Axes.arrow` method. 
	
	See the Matplotlib :meth:`Axes.arrow` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('arrow',) + args, 
			'kwargs': kwargs
		}
	)
	
def legend(*args, **kwargs):
	"""
	Displays a legend in the active axes of the active plot window. 
	All arguments are passed to the Matplotlib :meth:`Axes.legend` method. 
	
	See the Matplotlib :meth:`Axes.legend` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('legend',) + args, 
			'kwargs': kwargs
		}
	)

def xlabel(__xlabel, *args, **kwargs):
	"""
	Calling convention: xlabel(*xlabel*, ...)
	
	Labels the x-axis in the active axes of the active plot window with the 
	string given by *xlabel*. All remaining arguments are passed to the 
	Matplotlib :meth:`Axes.set_xlabel` method. 
	
	See the Matplotlib :meth:`Axes.set_xlabel` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('set_xlabel', __xlabel) + args, 
			'kwargs': kwargs
		}
	)

def ylabel(__ylabel, *args, **kwargs):
	"""
	Calling convention: ylabel(*ylabel*, ...)
	
	Labels the y-axis in the active axes of the active plot window with the 
	string given by *ylabel*. All remaining arguments are passed to the 
	Matplotlib :meth:`Axes.set_ylabel` method. 
	
	See the Matplotlib :meth:`Axes.set_ylabel` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('set_ylabel', __ylabel) + args, 
			'kwargs': kwargs
		}
	)

def title(__title, *args, **kwargs):
	"""
	Calling convention: title(*title*, ...)
	
	Sets the title for the active axes of the active plot window to *title*. 
	All remaining arguments are passed to the Matplotlib :meth:`Axes.set_title` 
	method. 
	
	See the Matplotlib :meth:`Axes.set_title` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('set_title', __title) + args, 
			'kwargs': kwargs
		}
	)

def text(*args, **kwargs):
	"""
	Calling convention: text(*x*, *y*, *s*, ...)
	
	Prints a text given by *s* at coordinates *x*, *y* in the active axes of 
	the active plot window. All remaining arguments are passed to the 
	Matplotlib :meth:`Axes.text` method. 
	
	See the Matplotlib :meth:`Axes.text` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('text', ) + args, 
			'kwargs': kwargs
		}
	)

def annotate(*args, **kwargs):
	"""
	Calling convention: annotate(*s*, *xy*, *xytext*, ...)
	
	Prints an annotation *s* at coordinates given by tuple *xytext* pointing to 
	a point with coordinates *xy* in the active axes of the active plot window. 
	All remaining arguments are passed to the Matplotlib :meth:`Axes.annotate` 
	method. 
	
	See the Matplotlib :meth:`Axes.annotate` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ('annotate', ) + args, 
			'kwargs': kwargs
		}
	)

def axes(__rect=(0.12,0.12,0.76,0.76), *args, **kwargs):
	"""
	Calling convention: axes(*rect*, ...)
	
	Creates new axes in the active plot window at the position and with size 
	given by the *rect* tuple (in relative coordinates). The tuple has the 
	following members: 
	
	* *left* - the left edge of the axes in relative canvas coordinates 
	  (0.0=left, 1.0=right)
	* *bottom* - the bottom edge of the axes in relative canvas coordinates
	  (0.0=bottom, 1.0=top)
	* *width* - the width of the axes in relative canvas coordinates
	* *height* - the height of the axes in relative canvas coordinates
	
	Defaults to (0.12, 0.12, 0.76, 0.76). 
	
	All remaining arguments are passed to the Matplotlib 
	:meth:`Figure.add_axes` method. 
	
	Returns the axes tag. If axes with the same specification already exist 
	the corresponding tag is returned without creating new axes. 
	
	See the Matplotlib :meth:`Figure.add_axes` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'axes'], 
			'args': (__rect,) + args, 
			'kwargs': kwargs
		}
	)

def subplot(__nv, __nh, __pos, *args, **kwargs):
	"""
	Calling convention: subplot(*nv*, *nh*, *pos*, ...)
	
	Creates new axes in the active plot window with a MATLAB style subplot 
	specification. The canvas is divided in *nv* rows of *nh* columns. The 
	individual cells are numbered starting with 1 (top left) to *nh* (top 
	right) and then continuing through the remaining rows to *nv* times *nh at 
	bottom right. *pos* specifies the number of the cell that is assigned to 
	the canvas. 
	
	All remaining arguments are passed to the Matplotlib 
	:meth:`Figure.add_subplot` method. 
	
	Returns the axes tag. If axes with the same specification already exist 
	the corresponding tag is returned without creating new axes. 
	
	See the Matplotlib :meth:`Figure.add_subplot` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'subplot'], 
			'args': (__nv, __nh, __pos) + args, 
			'kwargs': kwargs
		}
	)
	
def setAxes(tag):
	"""
	Activates axes tagged with *tag* in the active plot window. 
	
	Returns ``True`` on success. On failure all axes in the active plot window 
	become inactive. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'setaxes'], 
			'args': (tag,),  
			'kwargs': {}
		}
	)

def getAxes():
	"""
	Returns the tag corresponding to the active axes of the active plot window. 

	Returns ``None`` if no plot window is active or there are no active 
	axes in the active plot window. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'getaxes'], 
			'args': (),  
			'kwargs': {}
		}
	)

def getAllAxes():
	"""
	Returns the list of tags corresponding to all axes of the active plot 
	window. 

	Returns ``None`` if no plot window is active. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'getallaxes'], 
			'args': (),  
			'kwargs': {}
		}
	)

def deleteAxes(tag=None):
	"""
	Deletes the axes tagged with *tag* from the active plot window. 
	
	If *tag* is ommitted the active axes of the active plot window are deleted. 
	
	If the active axes are deleted no axes are made active and the active plot 
	window has no active axes. 
	
	Returns ``True`` on success.
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'delaxes'], 
			'args': (tag,),  
			'kwargs': {}
		}
	)

def clearAxes(tag=None):
	"""
	Clears the contents of the axes tagged with *tag* in the active plot 
	window. 
	
	If *tag* is ommitted the active axes are cleared. 
	
	Returns ``True`` on success. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'clearaxes'], 
			'args': ( tag, ), 
			'kwargs': {}
		}
	)

def hold(on=True):
	"""
	Sets the hold state for the active axes of the active plot window to *on*. 
	
	If the hold state is enabled (*on* is ``True``) adding a new trace does 
	not delete all previous traces from the axes. 
	
	See the Matplotlib :meth:`Axes.hold` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ( 'hold', on ), 
			'kwargs': {}
		}
	)

def setXAxisScale(*args, **kwargs):
	"""
	Calling convention: setXAxisScale(*type*, ...)
	
	Sets the type of the x-axis scale in the active axes of the active plot 
	window. Possible values of *type* are `linear`, `log`, and `symlog`. 
	
	All remaining arguments are passed to the Matplotlib 
	:meth:`Axes.set_xscale` method. 
	
	See the Matplotlib :meth:`Axes.set_xscale` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ( 'set_xscale', ) + args, 
			'kwargs': kwargs
		}
	)

def setYAxisScale(*args, **kwargs):
	"""
	Calling convention: setYAxisScale(*type*, ...)
	
	Sets the type of the y-axis scale in the active axes of the active plot 
	window. Possible values of *type* are `linear`, `log`, and `symlog`. 
	
	All remaining arguments are passed to the Matplotlib 
	:meth:`Axes.set_yscale` method. 
	
	See the Matplotlib :meth:`Axes.set_yscale` method for details. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'onaxes'], 
			'args': ( 'set_yscale', ) + args, 
			'kwargs': kwargs
		}
	)
	
def figure(tag=None, *args, **kwargs):
	"""
	Calling convention: figure(*tag*, *windowTitle*, *show*, *inFront*, 
	*figpx*, *dpi*, *figsize*, ...)
	
	Create a new plot window with tag given by *tag*. If a plot window with 
	the same tag exists it becomes inaccessible through its tag. If no *tag* 
	is given a tag is assigned automatically. 
	
	*windowTitle* specifies the title of the window. 
	
	If *show* is ``True`` (default) the window will be visible as soon as it is 
	created. 
	
	If *inFront* is ``True`` (default) the window will show up on top of all 
	other windows. 
	
	*figsize* is a tuple specifying the horizontal and vertical figure size in 
	inches. *dpi* is used for obtaining the size in pixels (for the screen). 
	
	*figpx* is a tuple specifying the horizontal and vertical figure size in 
	pixels. *dpi* is used for obtaining the figure size in inches (i.e. for 
	saving the figure in a Postscript file). When specified, *figpx* overrides 
	*figsize*. 
	
	All remaining arguments are passed to the constructor of the 
	:class:`pyopus.wxmplitf.PlotFrame` object which passes them on to the 
	constructor of the :class:`pyopus.wxmplitf.Plotpanel` object. 
	
	Returns the tag of the newly created plot window. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'new'], 
			'args': ( tag, ) + args, 
			'kwargs': kwargs
		}
	)

def setFigure(tag):
	"""
	Sets the active plot window to be the window tagged with *tag*. 
	
	Returns ``True`` on success. On failure all plot windows become inactive. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'set'], 
			'args': ( tag, ), 
			'kwargs': {}
		}
	)

def getFigure():
	"""
	Returns the tag of the active plot window. If no plot window is active 
	``None`` is returned. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'get'], 
			'args': (), 
			'kwargs': {}
		}
	)

def getAllFigures():
	"""
	Returns a list of tags of all plot windows. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'getall'], 
			'args': (), 
			'kwargs': {}
		}
	)

def closeFigure(tag=None):
	"""
	Closes a plot window tagged with *tag*. 
	If no *tag* is given, closes the active plot window. 
	
	Returns ``True`` on success. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'close'], 
			'args': ( tag, ), 
			'kwargs': {}
		}
	)
	
def closeAllFigures():
	"""
	Closes all plot windows. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'closeall'], 
			'args': (), 
			'kwargs': {}
		}
	)

def clearFigure(tag=None):
	"""
	Clears the contents of a plot window tagged with *tag*. 
	If no *tag* is given, clears the contents of teh active plot window. 
	
	Returns ``True`` on success. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'clear'], 
			'args': ( tag, ), 
			'kwargs': {}
		}
	)

def showFigure(on=True, tag=None):
	"""
	Shows (*on* is ``True``) or hides the plot window tagged with *tag*. 
	If *tag* is not given, shows/hides the active plot window. 
	
	Returns ``True`` on success. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'show'], 
			'args': ( on, tag ), 
			'kwargs': {}
		}
	)

def raiseFigure(tag=None):
	"""
	Raises the plot window tagged with *tag*. 
	If *tag* is not given, raises the active plot window. 
	
	Returns ``True`` on success. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'raise'], 
			'args': ( tag, ), 
			'kwargs': {}
		}
	)
	
def repaintFigure(on=True, tags=None):
	"""
	Enables (*on* is ``True``) or disables repainiting for the plot windows 
	tagged with *tag*. If no *tags* are given enables/disables repainting for 
	the active plot window. 
	
	Returns a list of boolean values where ``True`` indicates success. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'enablerepaint'], 
			'args': ( on, tags ), 
			'kwargs': {}
		}
	)

def saveFigure(fileName, tag=None):
	"""
	Saves the contents of a plot window tagged with *tag* to a file named 
	*fileName*. if no *tag* is given the contents of the active plot window 
	are saved. 
	
	See the :meth:`FigureCanvasWxAgg.print_figure` method. The available 
	file types are listed in the :attr:`FigureCanvasWxAgg.filetypes` 
	dictionary. 
		
	Returns ``True`` on success. 
	"""
	# Send full path (gui process may not have the same working directory). 
	return GUIctl.SendMessage({
			'cmd':['plot', 'savefigure'], 
			'args': ( os.path.realpath(fileName), tag ), 
			'kwargs': {}
		}
	)
	
def figureTitle(title):
	"""
	Sets the title of the active plot window's canvas to *title*. 
	
	This title appears on the top of the canvas. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'setfiguretitle'], 
			'args': (title, ), 
			'kwargs': {}
		}
	)

def windowTitle(title):
	"""
	Sets the title of the active plot window to *title*.
	
	This title appears in the title bar of the plot window. 
	"""
	return GUIctl.SendMessage({
			'cmd':['plot', 'setwindowtitle'], 
			'args': (title, ), 
			'kwargs': {}
		}
	)
	