"""
.. inheritance-diagram:: pyopus.wxmplplot.wxmgr
    :parts: 1
	
**Manager for Matplotlib plot windows**

Uses the :mod:`multiprocessing` module if it is available. If not, 
:mod:`threading` is used. 

The graphical part (wxPython + matplotlib) is running in a thread (or process 
if :mod:`multiprocessing` is used) and the part that issues the plotting 
commands (defined in the :mod:`~pyopus.wxmplplot.plotitf` module) runs in 
the main thread (or process). 

The main thread uses a :class:`GUIControl` object for sending and receiving 
messages from the graphical thread. The messages are sent and received using 
two :class:`Queue` objects. On the graphical thread's side a 
:class:`ControlApp` application is running. The main window of the application 
is a :class:`ControlWindow` object. Within the application there is a 
:class:`QueueMonitor` object watching the queue of incoming messages. 
The :class:`QueueMonitor` runs a separate thread with an infinite loop that 
picks up the incoming messages and dispatches them as :class:`ControlEvent` 
events to the :class:`ControlApp`. The :class:`ControlEvent` handler in 
:class:`ControlApp` calls the :meth:`ControlWindow.InterpretCommand` method 
that dispatches the message to the corresponding command handler. 
"""

import threading

# Use multiprocessing, if available. 
try:
	import multiprocessing
	use_multiprocessing=True
except:
	use_multiprocessing=False

import Queue
import cPickle
import traceback

from matplotlib import rcParams
from matplotlib.lines import Line2D

import wx 
from wxmplitf import *

__all__ = [ 'GUIfactory', 'GUIControl', 'GUIControlThreading', 
			'GUIControlMultiprocessing', 'GUIentry', 
			'ControlApp', 'ControlWindow', 'QueueMonitor', 
			'ControlEvent', 'EVT_CONTROL', 'EVT_CONTROL_ID', 'CONTROL_MSG',
			'use_multiprocessing' ]

#
# MessageCanvas providing a main window for messages
#

class ControlWindow(wx.Frame):
	"""
	This is the main GUI ControlWindow and command dispatcher. 
	
	*parent* is the parent window. *title* is the window title. 
	"""
	def __init__(self, parent, title):
		wx.Frame.__init__(self, parent, -1, title, pos=(150, 150), size=(350, 200))

		# Create the menubar
		menuBar = wx.MenuBar()

		# And a menu 
		menu = wx.Menu()
		
		# Add an item to the menu, using \tKeyName automatically
		# creates an accelerator, the third param is some help text
		# that will show up in the statusbar
		IDnewWin=menu.Append(-1, "&New plot\tCtrl-N", "New plot window")
		IDcloseAll=menu.Append(-1, "Close all plots", "Close all plot windows")
		menu.Append(wx.ID_EXIT, "E&xit\tAlt-X", "Exit")

		# Bind the menu event to an event handler
		self.Bind(wx.EVT_MENU, self.OnNewPlot, IDnewWin)
		self.Bind(wx.EVT_MENU, self.OnCloseAll, IDcloseAll)
		self.Bind(wx.EVT_MENU, self.OnExit, id=wx.ID_EXIT)

		# Put the menu on the menubar
		menuBar.Append(menu, "&File")
		self.SetMenuBar(menuBar)

		# Create staus bar
		self.CreateStatusBar()
		
		# Now create the Panel to put the other controls on.
		panel = wx.Panel(self)

		# Create message frame
		self.messages = wx.TextCtrl(panel, -1, style=wx.TE_MULTILINE|wx.TE_RICH2)
		self.messages.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.NORMAL))
		self.messages.SetSize(self.messages.GetBestSize())
		self.messages.SetEditable(False)
		self.OutputMessage("Control window ready.\n")

		# Use a sizer to layout the controls
		sizer = wx.BoxSizer(wx.VERTICAL)
		sizer.Add(self.messages, 1, wx.EXPAND)
		panel.SetSizer(sizer)
		panel.Layout()
		
		# Plot window storage
		self.plotWindows={}
		self.plotTagIndex={}
		
		# Active plot window
		self.activePlotWindowTag=None
		
		# Window counter
		self.plotWindowCounter=0
		
		# Initialize interpreter
		self.plot_cmdtab={}
		self.InitInterpreter()
		
		wx.EVT_WINDOW_DESTROY(self, self.OnDestroy)
		wx.EVT_CLOSE(self, self.OnClose)
		
	def InitInterpreter(self):
		"""
		Initializes the interpreter command tables. 
		
		Every command is a list of strings. The first string specifies the 
		command group and the second string the command. Currently only one 
		command group is available (``plot``). 
		
		Every command is passed an *args* tuple of positional arguments and a 
		*kwargs* dictionary of keyword arguments. 
		
		Every plot window is actually a 
		:class:`~pyopus.wxmplplot.wxmplitf.PlotFrame` object displaying a 
		Matplotlib :class:`Figure`. Every plot window is associated with a tag 
		at its creation. A tag is a hashable object (usually a string). 
		If not tag is specified at creation a tag is assigned automatically. 
		
		Every Matplotlib :class:`Figure` can have multiple :class:`Axes`. 
		Every axes represent a part of the figure where plotting takes place 
		within the coordinate system of the :class:`Axes`. Every axes in a 
		figure are associated with a numeric tag in a similar way as every plot 
		window has its tag. Axes tags are unique because they cannot be 
		assigned by the user. 
		
		Tags should be unique. If a duplicate figure tag occurs that older 
		figure with that tag becomes inaccessible. 
		
		The ``plot`` command group has the following commands:
		
		* ``exit`` - exits the interpreter
		* ``updatercparams`` - update the rc parameters with the given 
		  dictionary
		* ``new`` - create a new plot window and return its tag
		* ``set`` - activate a plot window
		* ``get`` - return the active plot window's tag
		* ``getall`` - return the list of tags coreresponding to all plot 
		  windows
		* ``close`` - close a plot window
		* ``closeall`` - close all plot windows
		* ``clear`` - clear contents of a plot window
		* ``show`` - show or hide aplot window
		* ``raise`` - raise a plot window
		* ``enablerepaint`` - enables/disables repainting of a plot window
		* ``setwindowtitle`` - sets the title of a plot window
		* ``setfiguretitle`` - sets the title of a figure (displayed on canvas)
		* ``setaxes`` - activate axes in the active plot window
		* ``getaxes`` - get the tag of the active axes in the active plot window
		* ``getallaxes`` - return the list of tags corresponding to all axes 
		  in the active plot window
		* ``axes`` - create new axes (from the relative position and size of 
		  axes within figure) and return the corresponding tag
		* ``subplot`` - create new axes (from the MATLAB style suplot 
		  specification) and return the corresponding tag
		* ``delaxes`` - removes axes from the active plot window
		* ``clearaxes`` - clears whatever is displayed within active axes of 
		  active plot window
		* ``savefigure`` - saves the contents of a figure to a file
		* ``onfigure`` - calls a method of the :class:`Figure` object 
		  corresponding to teh active plot window
		* ``onaxes`` - calls a method of the :class:`Axes` object corresponding 
		  to the active axes of the active figure
		  
		Note that the return values can be collected in the main thread 
		(process) only if talkback is enabled. 
		"""
		self.plot_cmdtab={
			'exit':				self.ExitInterpreter, 
			'updatercparams':	self.UpdateRCParams, 
			'new':				self.NewPlotWindow, 
			'set':				self.ActivatePlotWindow, 
			'get':				self.GetActivePlotWindow, 
			'getall':			self.GetAllPlotWindows, 
			'close':			self.ClosePlotWindow, 
			'closeall':			self.CloseAllPlotWindows, 
			'clear':			self.ClearPlotWindow, 
			'show':				self.ShowPlotWindow,
			'raise':			self.RaisePlotWindow, 
			'enablerepaint':	self.EnableRepaint, 
			'setwindowtitle':	self.SetPlotWindowTitle, 
			'setfiguretitle':	self.SetFigureTitle, 
			'setaxes':			self.ActivateAxes, 
			'getaxes':			self.GetActiveAxes, 
			'getallaxes':		self.GetAllAxes, 
			'axes':				self.NewAxes, 
			'subplot':			self.NewSubplot, 
			'delaxes':			self.DeleteAxes, 
			'clearaxes':		self.ClearAxes, 
			'savefigure':		self.SaveFigure, 
			'onfigure':			self.DoOnFigure, 
			'onaxes':			self.DoOnAxes
		}
		
	def ExitInterpreter(self):
		"""
		Handles the ``['plot', 'exit']`` command.
		
		Exits the interpreter (thread or process).
		"""
		self.Close()
	
	def UpdateRCParams(self, dict):
		"""
		Handles the ``['plot', 'updatercparams']`` command.
		
		Update the rcParams structure of Matplotlib with the given dictionary.
		Returns ``True`` on success. 
		"""
		try:
			rcParams.update(dict);
			retval=True
		except:
			retval=False
		
		return retval
		
	def ActivatePlotWindow(self, tag):
		"""
		Handles the ``['plot', 'set']`` command.
		
		Activates a plot window. Returns ``True`` on success.
		"""
		
		if tag in self.plotWindows:
			self.activePlotWindowTag=tag
			return True
		else:
			self.activePlotWindowTag=None
			return False
	
	def GetActivePlotWindow(self):
		"""
		Handles the ``['plot', 'get']`` command.
		
		Returns the tag of the active plot window.
		Returns ``None`` if no plot window is active. 
		"""
		
		return self.activePlotWindowTag
	
	def GetAllPlotWindows(self):
		"""
		Handles the ``['plot', 'getall']`` command.
		
		Returns the list of tags of all plot windows. 
		"""
		
		return self.plotWindows.keys() 
		
	def NewPlotWindow(self, tag=None, windowTitle="Untitled", show=True, inFront=True, **kwargs):
		"""
		Handles the ``['plot', 'new']`` command.
		
		*windowTitle* is the title string for the plot window. 
		
		If *show* is ``True`` the window becomes visible immediately after it 
		is created. 
		
		If *inFront* is ``True`` the window is created atop of all other 
		windows. 
		
		Creates a new plot window with tag given by *tag*. 
		If *tag* is ``None`` a tag for the window is generated automatically.  
		
		All remaining arguments are passed on to the constructor of the 
		:class:`~pyopus.wxmplplot.wxmplitf.PlotFrame` object. 
		
		Returns the plot window's tag or ``None`` on failure.
		"""
		
		if tag in self.plotWindows:
			return None
		else:
			if tag is None:
				tag=self.plotWindowCounter
			self.plotWindowCounter+=1
			
			window=PlotFrame(self, -1, str(tag)+": "+windowTitle, **kwargs)
			
			self.plotTagIndex[window]=tag
			self.plotWindows[tag]={}
			self.plotWindows[tag]['obj']=window
			self.plotWindows[tag]['titleArtist']=window.get_figure().suptitle('')
			self.plotWindows[tag]['axes']={}
			self.plotWindows[tag]['axesTagIndex']={}
			self.plotWindows[tag]['activeAxesTag']=None
			self.plotWindows[tag]['axndx']=0
			
			window.Connect(window.GetId(), -1, wx.wxEVT_CLOSE_WINDOW, self.OnChildClosed)
			EVT_CLOSEALL(window, window.GetId(), self.OnCloseAll)
			
			self.activePlotWindowTag=tag
		
			if show:
				window.Show(True)
			
			if inFront:
				window.Raise()
			
			return tag
			
	def _RemovePlotWindow(self, tag):
		"""
		Removes a plot window with givent *tag* from the list of plot windows. 
		
		Returns ``True`` on success.
		
		This should not be called directly. Call ClosePlotWindow() instead.
		"""
		
		if tag in self.plotWindows:
			window=self.plotWindows[tag]['obj']
		else:
			return False
		
		if window is not None:
			if tag is self.activePlotWindowTag:
				self.activePlotWindow=None
			
			del self.plotWindows[tag]
			del self.plotTagIndex[window]
			
			return True
		else:
			return False
		
	def _RemovePlotWindowByObject(self, window):
		"""
		Removes a plot window given by object *window* from the list of plot 
		windows. 
		
		Returns ``True`` on success.
		
		This should not be called directly. Call ClosePlotWindow() instead.
		"""
		
		if window in self.plotTagIndex:
			tag=self.plotTagIndex[window]
		else:
			return False
		
		if tag is not None:
			if tag is self.activePlotWindowTag:
				self.activePlotWindow=None
			
			del self.plotWindows[tag]
			del self.plotTagIndex[window]
			
			return True
		else:
			return False
		
	def ClosePlotWindow(self, tag=None):
		"""
		Handles the ``['plot', 'close']`` command.
		
		Closes a plot window with given *tag*. If *tag* is ``None`` closes the 
		active plot window.
		
		Returns ``True`` on success.
		"""
		
		if tag is None:
			tag=self.activePlotWindowTag
		
		if tag in self.plotWindows:
			self.plotWindows[tag]['obj'].Close()
		else:
			return False
			
	def CloseAllPlotWindows(self):
		"""
		Handles the ``['plot', 'closeall']`` command.
		
		Closes all plot windows.
		"""
		
		tags=self.plotWindows.keys()
		for tag in tags:
			self.plotWindows[tag]['obj'].Close()
	
	def EnableRepaint(self, on=True, tags=None):
		"""
		Handles the ``['plot', 'enablerepaint']`` command.
		
		Enables (*on* set to ``True``) or disables repainting for a list of 
		plot windows given by *tags*. 
		
		If *tags* is ``None`` the command is applied to the active plot window. 
		
		Returns ``True`` on success. 
		"""
		
		if tags is None:
			tags=[self.activePlotWindowTag]
			
		response=[]
		for tag in tags:
			if tag is not None and tag in self.plotWindows:
				window=self.plotWindows[tag]['obj']
				window.get_panel().set_repaint(on)
				if on:
					window.draw()
				response.append(True)
			else: 
				response.append(False)

		return response
	
	def ClearPlotWindow(self, tag=None):
		"""
		Handles the ``['plot', 'clear']`` command.
		
		Clears a plot window with given *tag*. If *tag* is ``None`` clears 
		the active plot window. 
		
		Returns ``True`` on success. 
		
		See the :meth:``Figure.clf`` method in Matplotlib. 
		"""
		
		if tag is None:
			tag=self.activePlotWindowTag
			
		if tag is not None:
			window=self.plotWindows[tag]['obj']
			fig=window.get_figure()
		else:
			return False
		
		# Remove all axes, set active axes to None, but leave counter untouched
		self.plotWindows[tag]['axes']={}
		self.plotWindows[tag]['axesTagIndex']={}
		self.plotWindows[tag]['activeAxesTag']=None
			
		fig.clf()
		window.draw();
		
		return True
	
	def ShowPlotWindow(self, on=True, tag=None):
		"""
		Handles the ``['plot', 'show']`` command.
		
		Shows (*on* set to ``True``) or hides a plot window. If *tag* is 
		``None`` the command is applied to the active plot window. 
		
		Returns ``True`` on success. 
		"""
		
		if tag is None:
			tag=self.activePlotWindowTag
			
		if tag is not None and tag in self.plotWindows:
			window=self.plotWindows[tag]['obj']
		else:
			return False
		
		window.Show(on)
		
		return True
	
	def RaisePlotWindow(self, tag=None):
		"""
		Handles the ``['plot', 'raise']`` command.
		
		Raises a plot window with given *tag*. If *tag* is ``None`` the command 
		is applied to the active plot window. 
		
		Returns ``True`` on success. 
		"""
		
		if tag is None:
			tag=self.activePlotWindowTag
			
		if tag is not None and tag in self.plotWindows:
			window=self.plotWindows[tag]['obj']
		else:
			return False
		
		window.Raise()
		
		return True
		
	def SetPlotWindowTitle(self, title):
		"""
		Handles the ``['plot', 'setwindowtitle']`` command.
		
		Sets the window title of the active plot window. 
		"""
		
		tag=self.activePlotWindowTag
		if tag is not None:
			window=self.plotWindows[tag]['obj']
		else:
			return False
		
		window.SetTitle(str(tag)+": "+title)
		
		return True
	
	def SetFigureTitle(self, title):
		"""
		Handles the ``['plot', 'setfiguretitle']`` command.
		
		Sets the window title of the active plot window. 
		"""
		
		tag=self.activePlotWindowTag
		if tag is not None:
			window=self.plotWindows[tag]['obj']
		else:
			return False
		
		artist=self.plotWindows[tag]['titleArtist'].set_text(title)
		window.draw()
		
		return True
		
	def ActivateAxes(self, axndx):
		"""
		Handles the ``['plot', 'setaxes']`` command.
		
		Activates axes tagged with *axndx* within the active plot window. 
		
		Returns ``True`` on success.
		"""
		
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
			fig=self.plotWindows[tag]['obj'].get_figure()
		else:
			return False
		
		if axndx in self.plotWindows[tag]['axes']:
			self.plotWindows[tag]['activeAxesTag']=axndx
			return True
		else:
			self.plotWindows[tag]['activeAxesTag']=None
			return False
	
	def GetActiveAxes(self):
		"""
		Handles the ``['plot', 'getaxes']`` command.
		
		Returns the tag of the active axes in the active plot window.
		
		Returns ``None`` if no plot window is active or there are no active 
		axes in the active plot window. 
		"""
		
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
			fig=self.plotWindows[tag]['obj'].get_figure()
		else:
			return None
		
		return self.plotWindows[tag]['activeAxesTag']
	
	def GetAllAxes(self):
		"""
		Handles the ``['plot', 'getallaxes']`` command.
		
		Returns the list of tags corresponding to all axes in the active plot 
		window. 
		
		Returns ``None`` if no plot window is active. 
		"""
		
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
			fig=self.plotWindows[tag]['obj'].get_figure()
		else:
			return None
		
		return self.plotWindows[tag]['axes'].keys()
		
	def NewAxes(self, *args, **kwargs):
		"""
		Handles the ``['plot', 'axes']`` command.
		
		Creates new axes in the active plot window from the relative position 
		and size of axes within figure. This method actually invokes the 
		:meth:`add_axes` method of a :class:`Figure`. The axes are tagged with 
		a unique numerical tag. The newly created axes become the active axes 
		of the plot window. 
		
		If axes with identical size and position are created twice, the second 
		time the existing axes become active and no additional axes are 
		created. 
		
		Returns the tag of the axes that become active or ``None`` on failure.
		"""
		
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
			window=self.plotWindows[tag]['obj']
			fig=window.get_figure()
		else:
			return None
		
		axes=fig.add_axes(*args, **kwargs)
		if axes is None:
			return None
		
		if axes in self.plotWindows[tag]['axesTagIndex']:
			# Old axes returned
			axndx=self.plotWindows[tag]['axesTagIndex'][axes]
		else:
			# New axes
			axndx=self.plotWindows[tag]['axndx']
			self.plotWindows[tag]['axndx']+=1
			self.plotWindows[tag]['axes'][axndx]=axes
			self.plotWindows[tag]['axesTagIndex'][axes]=axndx
		
		# Activate
		self.plotWindows[tag]['activeAxesTag']=axndx
		
		window.draw()
		
		return axndx
		
	def NewSubplot(self, *args, **kwargs):
		"""
		Handles the ``['plot', 'subplot']`` command.
		
		Creates new axes in the active plot window from the MATLAB style 
		subplot specification. This method actually invokes the 
		:meth:`add_subplot` method of a :class:`Figure`. The axes are tagged 
		with a unique numerical tag. The newly created axes become the active 
		axes of the plot window. 
		
		If axes with identical size and position are created twice, the second 
		time the existing axes become active and no additional axes are 
		created. 
		
		Returns the tag of the axes that become active or ``None`` on failure.
		"""
		
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
			window=self.plotWindows[tag]['obj']
			fig=window.get_figure()
		else:
			return None
		
		axes=fig.add_subplot(*args, **kwargs)
		if axes in self.plotWindows[tag]['axesTagIndex']:
			# Old axes returned
			axndx=self.plotWindows[tag]['axesTagIndex'][axes]
		else:
			# New axes
			axndx=self.plotWindows[tag]['axndx']
			self.plotWindows[tag]['axndx']+=1
			self.plotWindows[tag]['axes'][axndx]=axes
			self.plotWindows[tag]['axesTagIndex'][axes]=axndx
		
		# Activate
		self.plotWindows[tag]['activeAxesTag']=axndx
		
		window.draw()
		
		return axndx
	
	def DeleteAxes(self, axndx):
		"""
		Handles the ``['plot', 'delaxes']`` command.
		
		Deletes axes with numerical tag *axndx* from active figure. 
		
		If *axndx* is ``None`` deletes the active axes. 
		
		If active axes are deleted no axes are made active and the active plot 
		window has no active axes. 
		
		Returns ``True`` on success.
		
		See the :meth:`Figure.delaxes` method in Matplotlib. 
		"""
		
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
			window=self.plotWindows[tag]['obj']
			fig=window.get_figure()
		else:
			return False
		
		if axndx is None:
			axndx=self.plotWindows[tag]['activeAxesTag']
		
		if axndx in self.plotWindows[tag]['axes']:
			axes=self.plotWindows[tag]['axes'][axndx]
			del self.plotWindows[tag]['axes'][axndx]
			del self.plotWindows[tag]['axesTagIndex'][axes]
			if axndx is self.plotWindows[tag]['activeAxesTag']:
				plotWindows[tag]['activeAxesTag']=None
			fig.delaxes(axes)
			window.draw()
			
			return True
		else:
			return False
	
	def ClearAxes(self, axndx=None):
		"""
		Handles the ``['plot', 'clearaxes']`` command.
		
		Clears the contents of axes with numerical tag *axndx*. 
		
		If *axndx* is ``None`` the active axes are cleared. 
		
		Returns ``True`` on success. 
		
		See the :meth:`Axes.clear` method in Matplotlib. 
		"""
		
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
			window=self.plotWindows[tag]['obj']
			fig=window.get_figure()
		else:
			return False
		
		if axndx is None:
			axndx=self.plotWindows[tag]['activeAxesTag']
		
		if axndx is not None:
			axes=self.plotWindows[tag]['axes'][axndx]
			
			axes.clear()
			window.draw()
			
			return True
		else:
			return False
	
	def SaveFigure(self, fileName, tag=None):
		"""
		Handles the ``['plot', 'savefigure']`` command.
		
		Saves the contents of a plot window with given *tag* to a file with a 
		name given by *fileName*. File type is determined from the extension 
		in the *fileName*. 
		
		If *tag* is ``None`` the active plot window's contents are saved. 
		
		See the :meth:`FigureCanvasWxAgg.print_figure` method. The available 
		file types are listed in the :attr:`FigureCanvasWxAgg.filetypes` 
		dictionary. 
		
		Returns ``True`` on success. 
		"""
		
		# Default tag (active plot window)
		if tag is None:
			tag=self.activePlotWindowTag
		
		# Get window (PlotFrame)
		if tag is not None and tag in self.plotWindows:
			window=self.plotWindows[tag]['obj']
		else:
			return False
		
		# Get panel (PlotPanel)
		fig=window.get_panel()
		
		# Save to file
		fig.print_figure(fileName)
		
		return True
		
	def DoOnFigure(self, *args, **kwargs):
		"""
		Handles the ``['plot', 'onfigure']`` command.
		
		Runs a method of the :class:`Figure` object corresponding to the active 
		plot window. The method name is the first positional argument in the 
		*args* tuple. The remaining arguments in *args* and *kwargs* are 
		passed to the method. 
		
		Returns ``None`` on failure (no plot). 
		
		On success returns the method's return value. 
		"""
		
		# Get figure tag
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
		else:
			return None
			
		# Get window (PlotFrame) and its figure (PlotPanel)
		window=self.plotWindows[tag]['obj']
		fig=window.get_figure()
		
		# Run the command
		retval=(eval('fig.'+args[0]))(*(args[1:]), **kwargs)
		window.draw();
			
		return retval
	
	def DoOnAxes(self, *args, **kwargs):
		"""
		Handles the ``['plot', 'onaxes']`` command.
		
		Runs a method of the :class:`Axes` object corresponding to the active 
		axes of the active plot window. The method name is the first 
		positional argument in the *args* tuple. The remaining arguments in 
		*args* and *kwargs* are passed to the method. 
		
		Returns ``None`` on failure (no axes). 
		
		On success returns the method's return value. 
		"""
		
		# Get figure tag
		if self.activePlotWindowTag is not None:
			tag=self.activePlotWindowTag
		else:
			return None
			
		# Get window and its figure
		window=self.plotWindows[tag]['obj']
		fig=window.get_figure()
		
		# Get axes tag
		if self.plotWindows[tag]['activeAxesTag'] is not None:
			axndx=self.plotWindows[tag]['activeAxesTag']
		else:
			return None
		
		# Get axes
		axes=self.plotWindows[tag]['axes'][axndx]
		
		# Run the command
		retval=(eval('axes.'+args[0]))(*(args[1:]), **kwargs)
		
		# If picker not specified, use default picker with 10 pt tolerance
		if 'picker' not in kwargs:
			if type(retval) is not list:
				artists=[retval]
			else:
				artists=retval

			for artist in artists: 
				if isinstance(artist, (Line2D,)):
					artist.set_picker(10.0)
	
		window.draw();
			
		return retval
			
	def OutputMessage(self, msg, colour=(0,0,0), bgcolour=(255,255,255), fontfamily=wx.FONTFAMILY_SWISS):
		"""
		Displays a message given by *msg* in the control window. The message 
		colour and background colour are specified with the *colour* and 
		*bgcolour* arguments (tuples of RGB values between 0 and 255). 
		
		*fontfamily* specifies the font family for the message. 
		"""
		self.messages.SetInsertionPointEnd()
		style=self.messages.GetDefaultStyle()
		style.SetTextColour(wx.Colour(*colour))
		style.SetBackgroundColour(wx.Colour(*bgcolour))
		wxfont=style.GetFont()
		wxfont.SetFamily(fontfamily)
		style.SetFont(wxfont)
		self.messages.AppendText(msg)
			
	def InterpretCommand(self, command, args=[], kwargs={}):
		"""
		Interprets the command given by the *command* list of strings. The 
		first string is the command family and the second string is the 
		command. Currently only one command family is available (``plot``). 
		
		The arguments to the command are given by *args* and *kwargs*. 
		
		The command handlers are given in handler dictionaries with command 
		name for key. The handler dictionary for the ``plot`` command family 
		is is the :attr:`plot_cmdtab` member. 
		
		Every command is handled in a ``try-except`` block If an error occurs 
		during command execution the error is displayed in the command window. 
		
		The return value of the command handler is returned. 
		
		This method is invoked by the :class:`ControlApp` on every 
		:class:`ControlEvent`. The contents of the event specify the command, 
		the arguments and a talkback flag which is ``True`` if the return value 
		of the command handler should be sent back to the main thread 
		(process). 
		"""
		response=None
		
		if command[0]=='plot':
			# Plot commands
			# In future this will be a part of the plot controller object
			# Figure commands
			if command[1] in self.plot_cmdtab:
				try:
					# self.OutputMessage("Command: '"+str(command)+" "+str(args)+str(kwargs)+"'\n")
					response=self.plot_cmdtab[command[1]](*args, **kwargs)
					# self.OutputMessage("Response: '"+str(response)+"'\n")
				except:
					self.OutputMessage('Exception in '+str(command)+'\n', bgcolour=(255,200,200), colour=(0,0,0))
					self.OutputMessage(traceback.format_exc(), colour=(180,0,0))
					response=None
			else:
				self.OutputMessage("Unknown command [1]: '"+str(command)+"'\n")
		else:
			self.OutputMessage("Unknown command [0]: '"+str(command)+"'\n")
		
		return response
	
	def OnNewPlot(self, evt):
		"""
		Event handler for the menu option 'New plot'.
		"""

		self.NewPlotWindow()
	
	def OnCloseAll(self, evt):
		"""
		Event handler for the menu option 'Close all plots'.
		"""

		keys=self.CloseAllPlotWindows()
		
	def OnExit(self, evt):
		"""
		Event handler for the menu option 'Exit'.
		"""
		self.Close()
	
	def OnDestroy(self, evt):
		"""
		Event handler for EVT_WINDOW_DESTROY.
		"""
		pass
	
	def OnClose(self, evt):
		"""
		Event handler for EVT_CLOSE
		"""
		self.OnCloseAll(None)
		self.Destroy()
		
	def OnChildClosed(self, evt):
		"""
		Called when a child (plot window) receives a close event.
		"""
		
		# Remove it from the list
		self._RemovePlotWindowByObject(evt.GetEventObject())
		
		evt.Skip()


#
# ControlApp providing a matplotlib canvas in a top-level wxPython window
#

class ControlApp(wx.App):
	"""
	This is the GUI control application that will run in the GUI thread 
	(or process if the :mod:`multiprocessing` module is available). 
	
	The commands are received through the *commandQueue*. Responses are sent 
	back through the *responseQueue*. The *commandQueue* is monitored for 
	incoming messages by a :class:`QueueMonitor` object. 
	
	Any additional arguments are passed to teh constructor of :class:`wx.App`. 
	"""

	def __init__(self, title="GUI Control App", responseQueue=None, commandQueue=None, **kwds):
		self.controlWindow=None
		self.responseQueue=responseQueue
		self.commandQueue=commandQueue
		
		# Create a queue monitor that will pass the messages to us
		if self.commandQueue is not None:
			self.inputMonitor=QueueMonitor(self.commandQueue, self.ProcessMessage)
		else:
			self.inputMonitor=None
		
		wx.App.__init__(self, **kwds)

	def OnInit(self):
		"""
		Called automatically when the control GUI application starts. 
		
		Initializes the main GUI control window and makes it the top window. 
		Associates the :meth:`OnControlEvent` method with :class:`ControlEvent` 
		events. 

		This method starts to monitor the *commandQueue* with a 
		:class:`QueueMonitor` object. This object will post a 
		:class:`ControlEvent` event to the :class:`ControlApp` for every 
		received message.
		"""
		self.controlWindow=controlWindow=ControlWindow(None, "GUI Control Window")
		
		EVT_CONTROL(self, CONTROL_MSG, self.OnControlEvent)
		
		self.SetTopWindow(controlWindow)
		
		controlWindow.Raise()
		controlWindow.Show(True)
		
		# Start queue monitor
		if self.inputMonitor is not None:
			self.inputMonitor.run()
		
		return True
		
	def OnExit(self):
		"""
		Handles the wxPyton exit event by stopping the main loop of the 
		:class:`QueueMonitor` object. 
		"""
		if self.inputMonitor is not None:
			self.inputMonitor.halt()
			self.inputMonitor=None
			
		return True # wx.App.OnExit()
		
	def ProcessMessage(self, message, wantResponse=True): 
		"""
		This is the callback function that is invoked for every received 
		message. It posts a :class:`ControlEvent` event to :class:`ControlApp` 
		for every message received from the *commandQueue*. 
		
		If *wantResponse* is ``True`` the :class:`ControlEvent` requests 
		from its handler to send back the return value of the message handler. 
		
		Note that this method is invoked from the main loop's thread of the 
		:class:`QueueMonitor` object. 
		"""
		wx.PostEvent(self, ControlEvent(CONTROL_MSG, message, wantResponse))

	def OnControlEvent(self, evt):
		"""
		Handles a :class:`ControlEvent` event. The handler runs in the thread 
		in which the :class:`QueueMonitor` object was created. 
		
		For every event it calls the :meth:`InterpretCommand` method of the 
		:class:`ControlWindow` object. If response is requested the return 
		value of the :meth:`InterpretCommand` method is sent back by 
		pickling it and putting it in the *responseQueue*. 
		"""
		response=self.controlWindow.InterpretCommand(evt.message['cmd'], evt.message['args'], evt.message['kwargs'])
		if self.responseQueue and evt.wantResponse: 
			try:
				pr=cPickle.dumps(response)
			except:
				pr=cPickle.dumps(None)
			self.responseQueue.put(pr, True)
			

#
# Monitor a queue in a thread and post events to the wx event loop
#

class QueueMonitor(object):
	"""
	This is a simple queue monitor. It monitors *queue* for incoming messages 
	(Python objects) and invokes *callback* for every received message. 
	Every message must be a tuple. Its members are the arguments passed to 
	*callback*. 
	
	The main loop calls the :meth:`get` method of the queue and waits 
	indefinitely on a message. When a message is received, *callback* is
	invoked. The main loop executes in a separate thread so blocking does 
	not affect the thread in which the :class:`QueueMonitor` was created. 
	
	The main loop's thread is started with the :meth:`run` method. 
	"""
	def __init__(self, queue, callback):
		self.queue=queue
		self.callback=callback
	
	def mainLoop(self):
		"""
		The main loop. 
		"""
		while self.doLoop:
			try:
				msg=self.queue.get(True, 1)
				wx.CallAfter(self.callback, *msg)
			except: 
				pass
				
	def run(self):
		"""
		Starts the main loop's thread. 
		"""
		self.doLoop=True
		self.thread=threading.Thread(target=self.mainLoop)
		self.thread.start()
	
	def halt(self):
		"""
		Stops the main loop and exits its thread by setting a flag that signals
		the main loop to exit. 
		"""
		self.doLoop=False


#
# This is the EVT_CONTROL event. 
#

EVT_CONTROL_ID = wx.NewId()
"The wxPyhton identifier of a :class:`ControlEvent` event."

CONTROL_MSG = wx.NewId()
"The wxPyhton identifier of a message sender carried by a :class:`ControlEvent` event."

def EVT_CONTROL(win, id, func):
	"""
	Register to receive :class:`ControlEvent` events from a 
	:class:`ControlApp`.
	The events originating from *win* are to be handled by function *func*. 
	
	*id* is the identifier to be associated with the event. 
	"""
	win.Connect(id, -1, EVT_CONTROL_ID, func)

class ControlEvent(wx.PyCommandEvent):
	"""
	Control event for the ControlApp's main window. 
	
	The event carries a *message*. The response from the message receiver is 
	requested if *wantResponse* is ``True``. 
	
	*id* is the identifier associated with the event. 
	"""
	def __init__(self, id, message, wantResponse=True):
		wx.PyCommandEvent.__init__(self, EVT_CONTROL_ID, id)
		
		self.message=message
		self.wantResponse=wantResponse

	def Clone(self):
		"""
		Creates a clone of the :class:`ControlEvent` object. 
		"""
		return ControlEvent(self.GetId(), self.message, self.wantResponse)

#
# This are the GUIcontrol classes.  
#

def GUIentry(queueFromGUI, queueToGUI, redirect):
	"""
	Entry point of the GUI thread. 
	
	This function creates a :class:`ControlApp` object and starts the GUI 
	application's main loop. *queueFromGUI* and *queueToGUI* are two queues 
	that are used for communicating with the GUI. Commands are sent by writing 
	to *queueToGUI* and the responses are obtained by reading *queueFromGUI*. 
	
	If set to ``True`` *redirect* redirects ``stdout`` and ``stderr`` to a 
	window that pops up when needed. *redirect* is passed to the constructor of 
	:class:`wx.App`. 
	"""
	app=ControlApp(responseQueue=queueFromGUI, commandQueue=queueToGUI, redirect=redirect)
	app.MainLoop()
	# This is reached after main loop finishes

class GUIControl(object):
	"""
	This object is used in the main thread (or process) for sending the plot 
	commands to the graphical thread. It can also collect the value returned 
	by the functions called by the graphical thread for every command sent. 
	Sending back the return values (talkback) is enabled by default. 
	
	If *redirect* is ``True`` all output from ``stdout`` and ``stderr`` is 
	redirected to a window that pops up when needed. *redirect* is passed to 
	the constructor of :class:`wx.App`. 
	
	This is an abstract class. It lacks a queue for sending commands to the 
	graphical thread and a queue for receiving responses from the graphical 
	thread. 
	"""
	def __init__(self, redirect=False):
		self.redirect=redirect
		self.talkback=True
	
	def setTalkback(self, on):
		"""
		Enables (*on* set to ``True``) or disables the graphical thread 
		sending back the return values of the commands it executes. 
		"""
		self.talkback=on
		
	def CheckIfAlive(self):
		"""
		Returns ``True`` if the graphical thread (or process) is running. 
		"""
		return False
		
	def StartGUI(self):
		"""
		Starts the graphical thread (or process). 
		"""
		pass
	
	def SendMessage(self, message):
		"""
		Sends a *message* (Python object) to the graphical thread. 
		
		If talkback is enabled (see the :meth:`setTalkback` method) collects 
		the return value of the executed graphical command and returns it. 
		
		Returns ``None`` if talkback is disabled. This is also the case if 
		the graphical thread is not running. 
		"""
		if self.CheckIfAlive():
			# print "sending command", message['cmd']
			
			# Put message in the queue
			self.queueToGUI.put((message, self.talkback), True)

			if self.talkback:
				# print "  awaiting response"
				response=cPickle.loads(self.queueFromGUI.get(True))
				# print "    got it"
				return response
			else:
				return None
		else:
			return None
	
	def StopGUI(self):
		"""
		Stops the graphical thread by sending it the ``['plot', 'exit']`` 
		command. 
		"""
		self.SendMessage({
				'cmd':['plot', 'exit'], 
				'args': [], 
				'kwargs': {}
			}
		)

class GUIControlThreading(GUIControl):
	"""
	This object is used in the main thread for sending the plot 
	commands to the graphical thread. It is used when the 
	:mod:`multiprocessing` module is not available. 
	
	See the :class:`GUIControl` class for more information. 
	"""
	def __init__(self, redirect=False):
		GUIControl.__init__(self, redirect)
		self.queueFromGUI=Queue.Queue(-1)
		self.queueToGUI=Queue.Queue(-1)
		self.controlThread=None
	
	def CheckIfAlive(self):
		"""
		Returns ``True`` if the graphical thread is running. 
		"""
		if self.controlThread is not None: 
			if self.controlThread.isAlive():
				return True
			else: 
				self.controlThread=None
				return False
		else:
			return False
		
	def StartGUI(self):
		"""
		Starts the graphical thread. The thread entry point is the 
		:func:`GUIentry` function. 
		"""
		if not self.CheckIfAlive():
			self.controlThread=threading.Thread(target=GUIentry, args=(self.queueFromGUI, self.queueToGUI, self.redirect))
			self.controlThread.start()
			# We pass control to the rest of the program here. GUI thread is running now. 
		

class GUIControlMultiprocessing(GUIControl):
	"""
	This object is used in the main process for sending the plot 
	commands to the graphical process. It is used when the 
	:mod:`multiprocessing` module is available. 
	
	See the :class:`GUIControl` class for more information. 
	"""
	def __init__(self, redirect=False):
		GUIControl.__init__(self, redirect)
		self.queueFromGUI=None
		self.queueToGUI=None
		self.controlProcess=None
	
	def CheckIfAlive(self):
		"""
		Returns ``True`` if the graphical process is running. 
		"""
		if self.controlProcess is not None: 
			if self.controlProcess.is_alive():
				return True
			else: 
				self.controlProcess=None
				self.app=None
				return False
		else:
			return False
	
	def StartGUI(self):
		"""
		Starts the graphical process. The process entry point is the 
		:func:`GUIentry` function. 
		"""
		if not self.CheckIfAlive():
			self.queueFromGUI=multiprocessing.Queue(-1)
			self.queueToGUI=multiprocessing.Queue(-1)
		
			self.controlProcess=multiprocessing.Process(target=GUIentry, 
					args=(self.queueFromGUI, self.queueToGUI, self.redirect)
				)
			self.controlProcess.start()
			# We pass control to the rest of the program here. GUI thread is running now. 
			

#
# This is the factory for creating GUIcontrol objects
#
			
def GUIfactory(*args, **kwargs):
	"""
	This is a factory function that returns the appropriate :class:`GUIcontrol` 
	object depending on the availability of the :mod:`multiprocessing` module. 
	
	The arguments are passed on to the constructor of the :class:`GUIcontrol` 
	object. 
	"""
	global use_multiprocessing
	
	if use_multiprocessing:
		return GUIControlMultiprocessing(*args, **kwargs)
	else:
		return GUIControlThreading(*args, **kwargs)
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		